import 'package:cruiser/src/errors/param_not_provided.dart';
import 'package:cruiser/src/errors/param_not_registered.dart';
import 'package:cruiser/src/errors/route_not_found.dart';
import 'package:cruiser/src/logger/cruiser_logger.dart';
import 'package:cruiser/src/models/arguments_wrapper.dart';
import 'package:cruiser/src/models/cruiser_args.dart';
import 'package:cruiser/src/models/cruiser_options.dart';
import 'package:cruiser/src/models/cruiser_param.dart';
import 'package:cruiser/src/models/cruiser_route.dart';
import 'package:cruiser/src/models/param_map.dart';
import 'package:cruiser/src/navigator_observers/cruiser_stack_observer.dart';
import 'package:cruiser/src/transitions/cruiser_transition.dart';
import 'package:cruiser/src/transitions/custom_cruiser_transition.dart';
import 'package:cruiser/src/transitions/transition_factory.dart';
import 'package:cruiser/src/ui/page_not_found.dart';
import 'package:flutter/material.dart';

enum NavigationType { push, pushReplace, pushAndRemoveUntil, popAndPushNamed }

class Cruiser 
{
    Cruiser({this.options = const CruiserOptions(isLoggingEnabled: true),}) : assert(options != null) 
    {
        CruiserLogger.init(isLoggerEnabled: options.isLoggingEnabled);

        if (options != null && options.navigatorKey != null) 
        {
            this._navigatorKey = options.navigatorKey;
        } 
        else 
        {
            this._navigatorKey = GlobalKey<NavigatorState>();
        }
    }

    final CruiserOptions options;

    Map<String, CruiserRoute> _routeNameMappings = {};

    Map<String, Map<String, CruiserParam>> _routeParamsMappings = {};

    GlobalKey<NavigatorState> _navigatorKey;

    GlobalKey<NavigatorState> get navigatorKey => _navigatorKey;

    CruiserStackObserver _navigationStackObserver = CruiserStackObserver();

    CruiserStackObserver get navigationStackObserver => _navigationStackObserver;

    /// Get the registered routes names as a list.
    List<String> getRegisteredRouteNames() => _routeNameMappings.keys.toList();
    

    /// Retrieves the arguments passed in when calling the [navigate] function.
    ///
    /// Returned arguments are casted with the type provided, the type will always
    /// be a subtype of [CruiserArgs].
    ///
    /// Make sure to provide the appropriate type, that is, provide the same type
    /// as the one passed while calling [navigate], else a cast error will be
    /// thrown.
    static T args<T extends CruiserArgs>(BuildContext context) => (ModalRoute.of(context).settings.arguments as ArgumentsWrapper).cruiserArgs as T;
    

    static T param<T>(BuildContext context, String key) 
    {
        final routeSettings = ModalRoute.of(context).settings;
        final argumentsWrapper = (routeSettings.arguments as ArgumentsWrapper);
        final isParamNotRegistered = argumentsWrapper.routeParams == null || !argumentsWrapper.routeParams.containsKey(key);

        if (isParamNotRegistered) 
        {
            throw ParamNotRegisteredError(paramKey: key,routeName: routeSettings.name);
        }

        // Check for request paramter type with registered paramter type.
        final paramRegisterdType = argumentsWrapper.routeParams[key].paramType; // Type with which param was registed
        
        if (T != paramRegisterdType) 
        {
            CruiserLogger.instance
                ..warning("========================================")
                ..warning("Mismatching Paramter Type!")
                ..warning("Requested param '$key' with type '$T', but was declared with type '$paramRegisterdType'.\n")
                ..warning("Make sure to pass the type of variable which used when declaring the 'CruiserParam<T>'.")
                ..warning("========================================");
        }

        final defaultParamValue = argumentsWrapper.routeParams[key].defaultValue;
        final paramFromNavigationCall = argumentsWrapper.params != null ? argumentsWrapper.params[key] : null;
        return (paramFromNavigationCall ?? defaultParamValue) as T;
    }

    /// Add a new route to [Cruiser].
    ///
    /// Route is stored in [_routeNameMappings].
    ///
    /// If a route is provided with a name that was previously added, it will
    /// override the old one.
    void addRoute(CruiserRoute route) 
    {
        assert(route != null, "'route' argument cannot be null.");

        if (_routeNameMappings.containsKey(route.name)) 
        {
            CruiserLogger.instance.warning( "'${route.name}' has already been registered before. Overriding it!");
        }

        // Prepare route params
        final routeParams = <String, CruiserParam>{};

        if (route.params != null) 
        {
            route.params.forEach((cruiserParam) 
            {
                if (routeParams.containsKey(cruiserParam.name)) 
                {
                    CruiserLogger.instance.warning("'${cruiserParam.name}' param has already been specified for route $route. Overriding it!");
                }

                routeParams[cruiserParam.name] = cruiserParam;
            });
        }

        _routeNameMappings[route.name] = route;
        _routeParamsMappings[route.name] = routeParams;
    }

    /// Function used to navigate pages.
    ///
    /// [name] is the route name that was registered using [addRoute].
    ///
    /// [args] are optional arguments that can be passed to the next page.
    /// To retrieve these arguments use [args] method on [Cruiser].
    ///
    /// [navigationType] can be specified to choose from various navigation
    /// strategies such as [NavigationType.push], [NavigationType.pushReplace],
    /// [NavigationType.pushAndRemoveUntil].
    ///
    /// [removeUntilPredicate] should be provided if using
    /// [NavigationType.pushAndRemoveUntil] strategy.
    ///
    /// [transitions] is a list of transitions to be used when switching between
    /// pages. [transitionDuration] and [transitionCurve] are duration and curve
    /// used for these transitions.
    ///
    /// [params] are key pair values that can be passed when navigating to a route.
    ///
    /// Provide a custom transition with [customTransition] to cruiser instead of using
    /// inbuilt transitions, if not provided, cruiser will revert to use its default
    /// transitions or delegate to system's own transitions.
    Future<T> navigate<T>
    (
        String name, {
        CruiserArgs args,
        NavigationType navigationType = NavigationType.push,
        dynamic result,
        bool Function(Route<dynamic> route) removeUntilPredicate,
        List<CruiserTransition> transitions,
        Duration transitionDuration,
        Curve transitionCurve,
        Map<String, dynamic> params,
        CustomCruiserTransition customTransition,
    }){
        assert(name != null);
        assert(navigationType != null);
        assert(navigationType != NavigationType.pushAndRemoveUntil || removeUntilPredicate != null);

        _checkAndThrowRouteNotFound(name, args, navigationType);

        return _navigate
        (   
            name,
            args,
            navigationType,
            result,
            removeUntilPredicate,
            transitions,
            transitionDuration,
            transitionCurve,
            params,
            customTransition,
        )
        .then((value) => value as T);
    }

    /// Actual navigation is delegated by [navigate] method to this method.
    ///
    /// [name] is the route name that was registered using [addRoute].
    ///
    /// [args] are optional arguments that can be passed to the next page.
    /// To retrieve these arguments use [arguments] method on [Cruiser].
    ///
    /// [navigationType] can be specified to choose from various navigation
    /// strategies such as [NavigationType.push], [NavigationType.pushReplace],
    /// [NavigationType.pushAndRemoveUntil].
    ///
    /// [removeUntilPredicate] should be provided is using
    /// [NavigationType.pushAndRemoveUntil] strategy.
    ///
    /// [transitions] is a list of transitions to be used when switching between
    /// pages. [transitionDuration] and [transitionCurve] are duration and curve
    /// used for these transitions.
    ///
    /// [params] are key pair values that can be passed when navigating to a route.
    Future<dynamic> _navigate
    (
        String name,
        CruiserArgs args,
        NavigationType navigationType,
        dynamic result,
        bool Function(Route<dynamic> route) removeUntilPredicate,
        List<CruiserTransition> transitions,
        Duration transitionDuration,
        Curve transitionCurve,
        Map<String, dynamic> params,
        CustomCruiserTransition customTransition,
    ) async 
    {
        final routeParams = _routeParamsMappings[name];
        
        if (routeParams != null) 
        {
            routeParams.forEach((key, value) 
            {
                // Type of paramter passed should be the same
                // when type is declared.
                if (params.containsKey(value.name) && params[value.name] != null) 
                {
                    final passedParamType = params[value.name].runtimeType;

                    if (passedParamType != value.paramType) 
                    {
                        CruiserLogger.instance.warning("Invalid Parameter Type! '${value.name}' is declared with a type '${value.paramType}', but a '$passedParamType' was passed!");
                    }
                }

                // All paramters that are 'required' should be passed.
                bool isMissingRequiredParam = value.isRequired && (params == null || !params.containsKey(value.name));

                if (isMissingRequiredParam) 
                {
                    CruiserLogger.instance.warning(ParameterNotProvidedError(paramKey: value.name,routeName: name).toString());
                }
            });
        }

        final argsWrapper = ArgumentsWrapper(
            cruiserArgs: args,
            transitions: transitions,
            transitionDuration: transitionDuration,
            transitionCurve: transitionCurve,
            params: params,
            routeParams: _routeParamsMappings[name],
            customTransition: customTransition
        );

        // Evaluate if the route can be opend using route guard.
        final route = _routeNameMappings[name];

        switch (navigationType) 
        {
            case NavigationType.push:
            {
                return this.navigatorKey.currentState.pushNamed(name, arguments: argsWrapper);
            }
            case NavigationType.pushReplace: 
            {
                return this.navigatorKey.currentState.pushReplacementNamed(name, result: result, arguments: argsWrapper);
            }
            case NavigationType.pushAndRemoveUntil:
            {
                return this.navigatorKey.currentState.pushNamedAndRemoveUntil(name, removeUntilPredicate,arguments: argsWrapper);
            }
            case NavigationType.popAndPushNamed:
            {
                return this.navigatorKey.currentState.popAndPushNamed(name, result: result, arguments: argsWrapper);
            }
            default: return null;
        }
    }

    /// If the route is not registered throw an error
    /// Make sure to use the correct name while calling navigate.
    void _checkAndThrowRouteNotFound
    (
        String name,
        CruiserArgs args,
        NavigationType navigationType,
    ) 
    {
        assert(name != null);

        if (!_routeNameMappings.containsKey(name)) 
        {
            if (this.options.handleNameNotFoundUI) 
            {
                this.navigatorKey.currentState.push(MaterialPageRoute(builder: (BuildContext context) 
                {
                    return PageNotFound(routeName: name,args: args,navigationType: navigationType,);
                }));
            }

            throw RouteNotFoundError(name: name);
        }
    }

    /// Delegation for [Navigator.pop].
    void pop([dynamic result]) 
    {
        this.navigatorKey.currentState.pop(result);
    }

    /// Delegation for [Navigator.popUntil].
    void popUntil(void Function(Route<dynamic>) predicate)
    {
        this.navigatorKey.currentState.popUntil(predicate);
    }

    /// Generates the [RouteFactory] which builds a [Route] on request.
    ///
    /// These routes are built using the [CruiserRoute]s provided using
    /// [addRoute] method.
    RouteFactory generator() 
    {
        return (settings) 
        {
            final route = _routeNameMappings[settings.name];

            if (route == null) return null;

            ArgumentsWrapper argsWrapper = settings.arguments as ArgumentsWrapper;

        if (argsWrapper == null) 
        {
            argsWrapper = ArgumentsWrapper();
        }

        final CruiserArgs baseArgs = argsWrapper.cruiserArgs;

        // Select which transitions to use.
        // Priority:
        //   1. Transitions provided when route is called.
        //   2. Default transitions when route was registered.
        //   3. Default transition from CruiserOptions.
        final List<CruiserTransition> transitions = [];

        final bool areTransitionsProvidedInNavigate = argsWrapper.transitions != null && argsWrapper.transitions.isNotEmpty;
        final bool areTransitionsProvidedInRouteDeclaration = route.defaultTransitions != null && route.defaultTransitions.isNotEmpty;
        final bool areTransitionsProvidedInCruiserOptions = this.options.defaultTransitions != null;

        if (areTransitionsProvidedInNavigate) 
        {
            transitions.addAll(argsWrapper.transitions);
        } 
        else if (areTransitionsProvidedInRouteDeclaration) 
        {
            transitions.addAll(route.defaultTransitions);
        } 
        else if (areTransitionsProvidedInCruiserOptions) 
        {
            transitions.addAll(this.options.defaultTransitions);
        }

        final transitionDuration =  argsWrapper.transitionDuration ?? route.defaultTransitionDuration ?? this.options.defaultTransitionDuration;

        final transitionCurve = 
        argsWrapper.transitionCurve ?? route.defaultTransitionCurve ?? this.options.defaultTransitionCurve;

        final customTransition = argsWrapper.customTransition ?? route.customTransition ?? this.options.customTransition;

        bool shouldUseCustomTransition = customTransition != null;

        if (argsWrapper.customTransition != null) 
        {
            shouldUseCustomTransition = true;
        } 
        else if (areTransitionsProvidedInNavigate) 
        {
            shouldUseCustomTransition = false;
        } 
        else if (route.customTransition != null) 
        {
            shouldUseCustomTransition = true;
        } 
        else if (areTransitionsProvidedInRouteDeclaration) 
        {
            shouldUseCustomTransition = false;
        } 
        else if (this.options.customTransition != null) 
        {
            shouldUseCustomTransition = true;
        } 
        else if (areTransitionsProvidedInCruiserOptions)
        {
            shouldUseCustomTransition = false;
        }

        RouteSettings routeSettings = RouteSettings
        (
            name: settings.name,
            arguments: argsWrapper.copyWith(cruiserArgs: baseArgs != null ? baseArgs : route.defaultArgs)
        );

        return TransitionFactory.buildTransition
        (
            transitions: transitions,
            duration: transitionDuration,
            curve: transitionCurve,
            settings: routeSettings,
            customTransition: shouldUseCustomTransition ? customTransition : null,
            builder: (context) => route.builder(context,baseArgs ?? route.defaultArgs, ParamMap(route.name,argsWrapper.routeParams,argsWrapper.params)));
        };
    } 

    static RouteFactory unknownRouteGenerator() 
    {
        return (settings) 
        {
            return MaterialPageRoute
            (
                settings: settings,
                builder: (BuildContext context) 
                {
                    return PageNotFound(routeName: settings.name,args: settings.arguments);
                },
            );
        };
    }
}